'use strict';

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _assert = require('assert');

var _assert2 = _interopRequireDefault(_assert);

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

var _fs = require('fs');

var _fs2 = _interopRequireDefault(_fs);

var _pathExists = require('path-exists');

var _pathExists2 = _interopRequireDefault(_pathExists);

var _lodash = require('lodash.merge');

var _lodash2 = _interopRequireDefault(_lodash);

var _invariant = require('invariant');

var _invariant2 = _interopRequireDefault(_invariant);

var _babelCore = require('babel-core');

var babel = _interopRequireWildcard(_babelCore);

var _stripIndent = require('strip-indent');

var _stripIndent2 = _interopRequireDefault(_stripIndent);

var _commonTags = require('common-tags');

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _objectWithoutProperties(obj, keys) { var target = {}; for (var i in obj) { if (keys.indexOf(i) >= 0) continue; if (!Object.prototype.hasOwnProperty.call(obj, i)) continue; target[i] = obj[i]; } return target; }

module.exports = pluginTester;

var fullDefaultConfig = {
  babelOptions: {
    parserOpts: {},
    generatorOpts: {},
    babelrc: false
  }
};

function pluginTester() {
  var _ref = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

  var _ref$plugin = _ref.plugin,
      plugin = _ref$plugin === undefined ? requiredParam('plugin') : _ref$plugin,
      _ref$pluginName = _ref.pluginName,
      pluginName = _ref$pluginName === undefined ? getPluginName(plugin) : _ref$pluginName,
      _ref$title = _ref.title,
      describeBlockTitle = _ref$title === undefined ? pluginName : _ref$title,
      tests = _ref.tests,
      fixtures = _ref.fixtures,
      filename = _ref.filename,
      rest = _objectWithoutProperties(_ref, ['plugin', 'pluginName', 'title', 'tests', 'fixtures', 'filename']);

  if (fixtures) {
    testFixtures(_extends({
      plugin,
      pluginName,
      title: describeBlockTitle,
      fixtures,
      filename
    }, rest));
  }
  var testAsArray = toTestArray(tests);
  if (!testAsArray.length) {
    return;
  }
  var testerConfig = (0, _lodash2.default)({}, fullDefaultConfig, rest);

  describe(describeBlockTitle, function () {
    testAsArray.forEach(function (testConfig, index) {
      if (!testConfig) {
        return;
      }

      var _merge = (0, _lodash2.default)({}, testerConfig, toTestConfig({ testConfig, index, plugin, pluginName, filename })),
          skip = _merge.skip,
          only = _merge.only,
          title = _merge.title,
          code = _merge.code,
          babelOptions = _merge.babelOptions,
          output = _merge.output,
          snapshot = _merge.snapshot,
          error = _merge.error;

      (0, _assert2.default)(!skip && !only || skip !== only, 'Cannot enable both skip and only on a test');

      if (skip) {
        // eslint-disable-next-line jest/no-disabled-tests
        it.skip(title, tester);
      } else if (only) {
        // eslint-disable-next-line jest/no-focused-tests
        it.only(title, tester);
      } else {
        it(title, tester);
      }

      // eslint-disable-next-line complexity
      function tester() {
        (0, _invariant2.default)(code, _commonTags.oneLine`
            A string or object with a \`code\` or
            \`fixture\` property must be provided
          `);
        (0, _invariant2.default)(!babelOptions.babelrc || babelOptions.filename, 'babelrc set to true, but no filename specified in babelOptions');
        (0, _invariant2.default)(!snapshot || !output, '`output` cannot be provided with `snapshot: true`');

        var result = void 0;
        var errored = false;

        try {
          result = babel.transform(code, babelOptions).code.trim();
        } catch (err) {
          if (error) {
            errored = true;
            result = err;
          } else {
            throw err;
          }
        }

        var expectedToThrowButDidNot = error && !errored;
        (0, _assert2.default)(!expectedToThrowButDidNot, 'Expected to throw error, but it did not.');

        if (snapshot) {
          (0, _invariant2.default)(result !== code, _commonTags.oneLine`
              Code was unmodified but attempted to take a snapshot.
              If the code should not be modified, set \`snapshot: false\`
            `);
          var separator = '\n\n      ↓ ↓ ↓ ↓ ↓ ↓\n\n';
          var formattedOutput = [code, separator, result].join('');
          expect(`\n${formattedOutput}\n`).toMatchSnapshot(title);
        } else if (error) {
          assertError(error, result);
        } else if (output) {
          _assert2.default.equal(result, output, 'Output is incorrect.');
        } else {
          _assert2.default.equal(result, code, 'Expected output to not change, but it did');
        }
      }
    });
  });
}

function testFixtures(_ref2) {
  var plugin = _ref2.plugin,
      describeBlockTitle = _ref2.title,
      fixtures = _ref2.fixtures,
      filename = _ref2.filename,
      rest = _objectWithoutProperties(_ref2, ['plugin', 'title', 'fixtures', 'filename']);

  describe(`${describeBlockTitle} fixtures`, function () {
    var fixturesDir = getPath(filename, fixtures);
    _fs2.default.readdirSync(fixturesDir).forEach(function (caseName) {
      it(caseName.split('-').join(' '), function () {
        var fixtureDir = _path2.default.join(fixturesDir, caseName);
        var codePath = _path2.default.join(fixtureDir, 'code.js');
        var babelRcPath = _path2.default.join(fixtureDir, '.babelrc');

        var _merge2 = (0, _lodash2.default)({}, fullDefaultConfig, {
          babelOptions: {
            plugins: [plugin],
            // if they have a babelrc, then we'll let them use that
            // otherwise, we'll just use our simple config
            babelrc: _pathExists2.default.sync(babelRcPath)
          }
        }, rest),
            babelOptions = _merge2.babelOptions;

        var actual = babel.transformFileSync(codePath, babelOptions).code.trim();

        var output = _fs2.default.readFileSync(_path2.default.join(fixtureDir, 'output.js'), 'utf8').trim();

        _assert2.default.equal(output, actual, 'actual output does not match output.js');
      });
    });
  });
}

function toTestArray(tests) {
  tests = tests || []; // null/0/false are ok, so no default param
  if (Array.isArray(tests)) {
    return tests;
  }
  return Object.keys(tests).reduce(function (testsArray, key) {
    var value = tests[key];
    if (typeof value === 'string') {
      value = { code: value };
    }
    testsArray.push(_extends({
      title: key
    }, value));
    return testsArray;
  }, []);
}

function toTestConfig(_ref3) {
  var testConfig = _ref3.testConfig,
      index = _ref3.index,
      plugin = _ref3.plugin,
      pluginName = _ref3.pluginName,
      filename = _ref3.filename;

  if (typeof testConfig === 'string') {
    testConfig = { code: testConfig };
  }
  var _testConfig = testConfig,
      title = _testConfig.title,
      fixture = _testConfig.fixture,
      _testConfig$code = _testConfig.code,
      code = _testConfig$code === undefined ? getCode(filename, fixture) : _testConfig$code,
      _testConfig$fullTitle = _testConfig.fullTitle,
      fullTitle = _testConfig$fullTitle === undefined ? `${index + 1}. ${title || pluginName}` : _testConfig$fullTitle,
      _testConfig$output = _testConfig.output,
      output = _testConfig$output === undefined ? getCode(filename, testConfig.outputFixture) : _testConfig$output;

  return (0, _lodash2.default)({
    babelOptions: { filename: getPath(filename, fixture) }
  }, testConfig, {
    babelOptions: { plugins: [plugin] },
    title: fullTitle,
    code: (0, _stripIndent2.default)(code).trim(),
    output: (0, _stripIndent2.default)(output).trim()
  });
}

function getCode(filename, fixture) {
  if (!fixture) {
    return '';
  }
  return _fs2.default.readFileSync(getPath(filename, fixture), 'utf8');
}

function getPath(filename, basename) {
  if (!basename) {
    return undefined;
  }
  if (_path2.default.isAbsolute(basename)) {
    return basename;
  }
  return _path2.default.join(_path2.default.dirname(filename), basename);
}

// eslint-disable-next-line complexity
function assertError(error, result) {
  if (typeof error === 'function') {
    if (!(result instanceof error || error(result) === true)) {
      throw result;
    }
  } else if (typeof error === 'string') {
    _assert2.default.equal(result.message, error, 'Error message is incorrect');
  } else if (error instanceof RegExp) {
    (0, _assert2.default)(error.test(result.message), `Expected ${result.message} to match ${error}`);
  } else {
    (0, _invariant2.default)(typeof error === 'boolean', 'The given `error` must be a function, string, boolean, or RegExp');
  }
}

function requiredParam(name) {
  throw new Error(`${name} is a required parameter.`);
}

function getPluginName(plugin) {
  var name = void 0;
  try {
    name = plugin(babel).name;
  } catch (error) {
    // eslint-disable-next-line no-console
    console.error(_commonTags.oneLine`
        Attempting to infer the name of your plugin failed.
        Tried to invoke the plugin which threw the error.
      `);
    throw error;
  }
  (0, _invariant2.default)(name, 'The `pluginName` must be inferable or provided.');
  return name;
}